#include "foo_uie_tabs.h"
#include "config.h"
//#include "dragdrop.h"


static const int WINDOW_BORDER_SIZE_X = GetSystemMetrics(SM_CXBORDER);
static const int WINDOW_BORDER_SIZE_Y = GetSystemMetrics(SM_CYBORDER);

#define FOO_UIE_TABS_VERSION "0.2.6"

#ifdef _DEBUG
#define FOO_UIE_TABS_DEBUG " (debug)"
#else
#define FOO_UIE_TABS_DEBUG
#endif

DECLARE_COMPONENT_VERSION("Tabbed panel modified"FOO_UIE_TABS_DEBUG,FOO_UIE_TABS_VERSION,
						  "Tabbed ui extension panel modified "FOO_UIE_TABS_VERSION"\n\nA tabbed window you can add your ui extensions to.\nRequires foo_ui_columns.dll v0.1.3b1 +");

static window_factory<uie_tabs_panel> g_tabs_ext;
static window_host_factory<uieh_tabs> g_tabs_host;


// remember to delete objects when done
void get_available_extension_infos(ui_extension::window_info_list_simple & out, DWORD type, ui_extension::window_host_ptr & host);

uieh_tabs::tabbed_ext::tabbed_ext(ui_extension::window_ptr &p, const char * name) : ptr(p)
{
	if (name)
		caption.set_string(name);
	else if(!p->get_short_name(caption))
		p->get_name(caption);
}

uieh_tabs::tabbed_ext::~tabbed_ext() {  ptr.release(); }

bool uieh_tabs::tabbed_ext::has_default_name()
{
	if(ptr.is_valid())
	{
		string8 name;
		if(!ptr->get_short_name(name))
			ptr->get_name(name);
		if (!uStringCompare(name,caption)) return true;
	}
	return false;
}

uieh_tabs::uieh_tabs(uie_tabs_panel * parent) : wnd(0), tab_wnd(0), active_wnd(0), selected_tab((unsigned)-1),selected_tab_cfg((unsigned)-1), parent_ext(parent), font(0), tab_orient((unsigned)-1), hide_tab(false), limit_min_height(cfg_min_height)
{}

uieh_tabs::uieh_tabs() : wnd(0), tab_wnd(0), active_wnd(0), selected_tab((unsigned)-1),selected_tab_cfg((unsigned)-1), parent_ext(0), font(0), tab_orient((unsigned)-1), hide_tab(false), limit_min_height(cfg_min_height)
{}

uieh_tabs::~uieh_tabs() { exts.delete_all(); pre_exts.delete_all();}

pfc::ptr_list_t<uieh_tabs> uieh_tabs::g_list_tabs;

static bool is_extension_wnd_valid(HWND wnd)
{
	if (!wnd) return false;
	UINT style = uGetWindowLong(wnd,GWL_STYLE);
	UINT ex_style = uGetWindowLong(wnd,GWL_EXSTYLE);
	if (!(style & WS_CHILD))
		return false;
	if (style & (WS_CAPTION | WS_BORDER | WS_THICKFRAME | DS_MODALFRAME))
		return false;
	return true;
}

void uieh_tabs::insert_extension (ui_extension::window_ptr & ext, const char * title)
{
	if(!ext.is_valid())
		return;
	else if(!wnd){
		pre_exts.add_item(new tabbed_ext(ext,title));
		return;
	}
	HWND ext_wnd = ext->create_or_transfer_window(wnd, 	this/*ui_extension::window_host_ptr(this)*/);

	if (!is_extension_wnd_valid(ext_wnd))
	{ ext->destroy_window(); return; }

	tabbed_ext * te;
	te = new tabbed_ext(ext,title);

	uTCITEM item;
	memset(&item,0,sizeof(item));
	item.mask = TCIF_PARAM|TCIF_TEXT;
	item.lParam = reinterpret_cast<long>(te);
	item.pszText = const_cast<char*>(te->caption.get_ptr());

	int idx = TabCtrl_GetItemCount(tab_wnd);
	if (uTabCtrl_InsertItem(tab_wnd,idx,&item) != -1)
	{	
		ShowWindow(ext_wnd, SW_SHOW);
		TabCtrl_SetCurSel(tab_wnd,idx);
		exts.add_item(te);
		on_tab_sel_change();
		InvalidateRect(tab_wnd,NULL,TRUE);
	}
	else
	{
		te->ptr->destroy_window();
		delete te;
	}
}

void uieh_tabs::insert_extension (const GUID & in, const char * title)
{
	ui_extension::window_ptr ext;
	if(ui_extension::window::create_by_guid(in, ext))
	{
		insert_extension (ext,title);
		ext.release();
	}
}

void uieh_tabs::insert_extension (const GUID & in, unsigned height, unsigned width)
{
	insert_extension(in);
}

void uieh_tabs::insert_extension (ui_extension::window_ptr & p_ext, unsigned height, unsigned width)
{
	insert_extension(p_ext);
}

void uieh_tabs:: insert_extension(const GUID & in, stream_reader * p_reader, t_size p_size, abort_callback & p_abort, const char * title)
{
	ui_extension::window_ptr ext;
	if(ui_extension::window::create_by_guid(in, ext))
	{
		if(p_size > 0)
			ext->set_config(p_reader, p_size, p_abort);
		if(wnd){
			insert_extension (ext,title);
		} else{
			pre_exts.add_item(new tabbed_ext(ext,title));
		}
		ext.release();
	} else{
		if(p_size > 0)
			p_reader->skip(p_size, p_abort);
	}
}

void uieh_tabs::relinquish_ownership(HWND wnd)
{
	TCITEM item;
	memset(&item,0,sizeof(item));
	item.mask = TCIF_PARAM;

	int n = TabCtrl_GetItemCount(tab_wnd);
	for (int i = 0; i < n; i++)
	{
		TabCtrl_GetItem (tab_wnd,i,&item);
		ui_extension::window_ptr ext = 
			reinterpret_cast<tabbed_ext*>(item.lParam)->ptr;
		if (ext.is_valid() && ext->get_wnd()==wnd)
		{
			delete_tab(i,false);
			break;
		}
	}
}

HWND uieh_tabs::create_window(HWND parent)
{
	wnd = uCreateDialog(IDD_PANEL,parent,dialog_proc,reinterpret_cast<long>(this));
	set_tab_bar_orientation(static_cast<tab_orientation>(tab_orient));
	tab_orient = (unsigned)-1;

	for(t_size i=0; i< pre_exts.get_count();i++){
		insert_extension(pre_exts[i]->ptr, pre_exts[i]->caption);
	}
	pre_exts.delete_all();
	if (selected_tab_cfg < exts.get_count())
	{
		TabCtrl_SetCurSel(tab_wnd, selected_tab_cfg);
		on_tab_sel_change();
	}
	selected_tab_cfg = (unsigned)-1;
    g_list_tabs.add_item(this);
    if (hide_tab) {
        ShowWindow(tab_wnd, SW_HIDE);
    }
	return wnd;
}

void uieh_tabs::destroy_window()
{
	for (unsigned i = 0; i < exts.get_count(); i++){
		TRACK_CODE("destroy child panel", exts[i]->ptr->destroy_window());
	}
	if(wnd){
		TRACK_CODE("destroy host", DestroyWindow(wnd));
	}
    g_list_tabs.remove_item(this);
	wnd=0;
	exts.delete_all();
	pre_exts.delete_all();
}

BOOL CALLBACK uieh_tabs::dialog_proc(HWND wnd, UINT msg, WPARAM wp, LPARAM lp)
{
	if (msg==WM_INITDIALOG)
		uSetWindowLong(wnd,GWL_USERDATA,lp);

	uieh_tabs * ext = 
		reinterpret_cast<uieh_tabs*>(uGetWindowLong(wnd,GWL_USERDATA));
	if (ext){
		return ext->on_message(wnd,msg,wp,lp);
	}
	return 0;
}

BOOL CALLBACK uieh_tabs::on_message(HWND wnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		tab_wnd = GetDlgItem(wnd,IDC_TAB);
		TabCtrl_SetPadding(GetDlgItem(wnd,IDC_TAB),(int)cfg_tab_pad_width,(int)cfg_tab_pad_height);
		REGISTER_PADDING_NOTIFY(wnd);
//		g_notify_min_height.add_window(wnd);
//		g_notify_font.add_window(wnd);
		font = CreateFontIndirect(&(LOGFONT)cfg_font);
		uSendDlgItemMessage(wnd,IDC_TAB,WM_SETFONT,(WPARAM)font,MAKELPARAM(TRUE,0));

		RECT rc;
		GetClientRect(wnd,&rc);
		if (get_tab_bar_orientation() == TABS_ON_TOP)
			rc.top += 2*WINDOW_BORDER_SIZE_Y;
		SetWindowPos(
			tab_wnd, NULL, 
			rc.left,rc.top,
			0,0,
			SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE
			);
		return 1;
	case WM_DESTROY:
		if(font){		
			DeleteObject(font);
		}
		font = 0;
		tab_wnd = 0;
		UNREGISTER_PADDING_NOTIFY(wnd);
//		g_notify_min_height.remove_window(wnd);
//		g_notify_font.remove_window(wnd);
		break;
	case WM_KEYDOWN:
		{
			service_ptr_t<keyboard_shortcut_manager> p_keyboard;
		if (get_keyboard_shortcuts_enabled()
			&& keyboard_shortcut_manager::g_get(p_keyboard)
			&& p_keyboard->on_keydown(keyboard_shortcut_manager::TYPE_MAIN,wp))
			break;
		else if (wp == VK_TAB)
			ui_extension::window::g_on_tab(wnd);
		break;
		}
	case WM_CONTEXTMENU:
		{
			POINT pt = {(short)LOWORD(lp), (short)HIWORD(lp)};
			int target = -1;
			TCHITTESTINFO info;
			memset(&info,0,sizeof(info));
			info.pt = pt;
			ScreenToClient(tab_wnd,&info.pt);
			int idx = TabCtrl_HitTest(tab_wnd,&info);
			if (info.flags & TCHT_ONITEM)
				target = idx;
			if (pt.x == -1 && pt.y == -1)
			{
				RECT r;
				int focus = TabCtrl_GetCurFocus(tab_wnd);
				if (focus != -1)
				{
					TabCtrl_GetItemRect(tab_wnd,focus,&r);
					pt.x = r.right-10;
					pt.y = r.bottom-10;
					ClientToScreen(tab_wnd,&pt);
				}
				else
				{
					GetWindowRect(wnd,&r);
					pt.x = r.left;
					pt.y = r.top;
				}
			}

			if (target == -1)
			{
				RECT r = get_display_area();
				MapWindowPoints(wnd,NULL,(LPPOINT)&r,2);
				if (active_wnd && PtInRect(&r,pt))
					target = selected_tab;
			}
			TRACK_CODE("context" , do_context_menu(pt,target));
		}
		return 1;
	case WM_SIZE:
		{
			RECT rc;
			GetClientRect(wnd,&rc);
			if (!IsRectEmpty(&rc)) 
			{
				SetWindowPos(
					tab_wnd, NULL, 
					0,0,
					rc.right-rc.left, rc.bottom-rc.top-2*WINDOW_BORDER_SIZE_Y, 
					SWP_NOMOVE|SWP_NOZORDER|SWP_NOACTIVATE
					);
				if (active_wnd)
					size_child_window(active_wnd);
			}
		}
		break;
	case WM_MOVE:
		{
			RECT rc;
			GetClientRect(wnd,&rc);
			if (get_tab_bar_orientation() == TABS_ON_TOP)
				rc.top += 2*WINDOW_BORDER_SIZE_Y;
			SetWindowPos(
				tab_wnd, NULL, 
				rc.left,rc.top,
				0,0,
				SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE
				);
		}
		break;
	case WM_NOTIFY:
		switch (((LPNMHDR)lp)->code)
		{
		case TCN_SELCHANGE:
			on_tab_sel_change();
			break;
		}
		break;
#if 0 // hrrrmm
	case WM_PAINT:
		if (active_wnd)
			InvalidateRect(active_wnd,NULL,TRUE);
		break;
#endif
	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO mmi = LPMINMAXINFO(lp);
            if (limit_min_height) {
                mmi->ptMinTrackSize.y = limit_min_height;
            }
            /*else if (cfg_min_height) {
    			mmi->ptMinTrackSize.y = cfg_min_height;
            }*/
		}
		break;
	case MIN_HEIGHT_CHANGED:
		if (parent_ext && parent_ext->get_host().is_valid())
			parent_ext->get_host()->on_size_limit_change(wnd,ui_extension::size_limit_minimum_height);
		break;
	case TAB_PADDING_CHANGED:
		TabCtrl_SetPadding(GetDlgItem(wnd,IDC_TAB),(int)cfg_tab_pad_width,(int)cfg_tab_pad_height);
		// hack to redraw the tabs at the new size
		if (exts.get_count()>0)
			rename_tab(0,exts[0]->caption);
		break;
	case FONT_CHANGED:
		{
			if (font) DeleteObject(font);
			font = CreateFontIndirect(&(LOGFONT)cfg_font);
			uSendDlgItemMessage(wnd,IDC_TAB,WM_SETFONT,(WPARAM)font,MAKELPARAM(TRUE,0));
			if (active_wnd)
				size_child_window(active_wnd);
		}
		break;
	case WM_COMMAND:
		switch(wp)
		{
		case IDOK:
		case IDCANCEL:
			if (active_wnd) uSendMessage(active_wnd,msg,wp,lp);
			break;
		}
		break;
	}
	return 0;
}
//#define TAB_STYLE_CHANGE
void uieh_tabs::do_context_menu(POINT pt, unsigned target_tab)
{
	enum { 
		ID_CLOSE=1, ID_RENAME,
		ID_LEFT, ID_RIGHT,
		ID_TAB_TOP, ID_TAB_BOTTOM, ID_TAB_LEFT,
#ifdef TAB_STYLE_CHANGE
		ID_TAB_TAB, ID_TAB_BTN,
#endif
        ID_HIDETAB, ID_LIMIT_MIN_HEIGHT,
		ID_SETTINGS, //ID_CLOSEPANEL,
		ID_OTHER
	};
	HMENU menu = CreatePopupMenu();
	HMENU panels = CreateMenu();
	HMENU tab_position = CreateMenu();
#ifdef TAB_STYLE_CHANGE
	HMENU tab_menu = CreateMenu();
#endif
	uAppendMenu(menu,MF_POPUP|MF_STRING,(UINT)panels,"&Add tab");
	uAppendMenu(menu,MF_POPUP|MF_STRING,(UINT)tab_position,"Tab-bar &position");
#ifdef TAB_STYLE_CHANGE
	uAppendMenu(menu,MF_POPUP|MF_STRING,(UINT)tab_menu, "&Tab-style");
#endif
    if (hide_tab) {
        uAppendMenu(menu,MF_STRING|MF_CHECKED,ID_HIDETAB,"&Hide Tabbar");
    } else {
        uAppendMenu(menu,MF_STRING,ID_HIDETAB,"&Hide Tabbar");
    }
    uAppendMenu(menu, MF_STRING, ID_LIMIT_MIN_HEIGHT, "&Limit MinHeight...");
	uAppendMenu(menu,MF_STRING,ID_SETTINGS,"&Settings...");
//	uAppendMenu(menu,MF_STRING,ID_CLOSEPANEL,"&Close panel stack");
	
	if (get_tab_bar_orientation() == TABS_ON_TOP)
	{
		uAppendMenu(tab_position,MF_STRING|MF_CHECKED|MF_DISABLED,ID_TAB_TOP,"&Top");
		uAppendMenu(tab_position,MF_STRING,ID_TAB_BOTTOM,"&Bottom");
//		uAppendMenu(tab_position,MF_STRING,ID_TAB_LEFT,"&Left");
	}
	else if(get_tab_bar_orientation() == TABS_ON_BOTTOM)
	{
		uAppendMenu(tab_position,MF_STRING,ID_TAB_TOP,"&Top");
		uAppendMenu(tab_position,MF_STRING|MF_CHECKED|MF_DISABLED,ID_TAB_BOTTOM,"&Bottom");
//		uAppendMenu(tab_position,MF_STRING,ID_TAB_LEFT,"&Left");
	}
	else{
		uAppendMenu(tab_position,MF_STRING,ID_TAB_TOP,"&Top");
		uAppendMenu(tab_position,MF_STRING,ID_TAB_BOTTOM,"&Bottom");
//		uAppendMenu(tab_position,MF_STRING|MF_CHECKED|MF_DISABLED,ID_TAB_LEFT,"&Left");
	}
#ifdef TAB_STYLE_CHANGE	
	if (get_tab_style() == TABS_TAB)
	{
		uAppendMenu(tab_menu,MF_STRING|MF_CHECKED|MF_DISABLED,ID_TAB_TAB,"&Tab");
		uAppendMenu(tab_menu,MF_STRING,ID_TAB_BTN,"&Button");
	}
	else {
		uAppendMenu(tab_menu,MF_STRING,ID_TAB_TAB,"&Tab");
		uAppendMenu(tab_menu,MF_STRING|MF_CHECKED|MF_DISABLED,ID_TAB_BTN,"&Button");
	}
#endif
	int id = ID_OTHER;
	ui_extension::window_info_list_simple extensions;
	ui_extension::window_host_ptr tmp_host(this);
	TRACK_CODE("get extensions", get_available_extension_infos(extensions, uie::type_panel|uie::type_toolbar|uie::type_layout|uie::type_playlist|uie::type_splitter, tmp_host));
	extensions.sort();
	int n = extensions.get_count();
	string8 category;
	HMENU cat_menu = 0;
	for (int i = 0; i < n; i++)
	{
		ui_extension::window_info_simple & ext = extensions[i];
		if (uStringCompare(category,ext.category))
		{
			category = ext.category;
			cat_menu = CreateMenu();
			uAppendMenu(panels,MF_POPUP|MF_STRING,(UINT)cat_menu,category.get_ptr());
		}
		uAppendMenu(cat_menu,MF_STRING,id++,ext.name.get_ptr());
	}
	
	const int ID_EXT_MENU_BASE = ID_OTHER+id;
	ui_extension::window_ptr ext = 0;
	void * ext_user_data = 0;
	ui_extension::menu_hook_impl menu_hook;
	if (target_tab != -1)
	{
		DWORD move_l_flags = MF_STRING;
		DWORD move_r_flags = MF_STRING;
		if (target_tab == 0)
			move_l_flags |= MF_GRAYED;
		if (target_tab == (unsigned)TabCtrl_GetItemCount(tab_wnd)-1)
			move_r_flags |= MF_GRAYED;
		
		uInsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,"");
		uInsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_CLOSE,"Cl&ose tab");
		uInsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_RENAME,"Re&name tab...");
		uInsertMenu(menu,0,MF_BYPOSITION|move_r_flags,ID_RIGHT,"Move &right");
		uInsertMenu(menu,0,MF_BYPOSITION|move_l_flags,ID_LEFT,"Move &left");
#ifndef _DEBUG
		{
			TCITEM item;
			memset(&item,0,sizeof(item));
			item.mask = TCIF_PARAM;
			TabCtrl_GetItem (tab_wnd,target_tab,&item);
			ext = (reinterpret_cast<tabbed_ext*>(item.lParam))->ptr;
			if (ext.is_valid()){
				TRACK_CODE("get child menu", ext->get_menu_items(menu_hook));
				if(menu_hook.get_children_count()){
					uAppendMenu(menu,MF_SEPARATOR,ID_EXT_MENU_BASE,"");
					TRACK_CODE("build menu", menu_hook.win32_build_menu(menu, ID_EXT_MENU_BASE + 1, 100));
				}
//				ext->build_menu(menu,ID_EXT_MENU_BASE,pt,false,ext_user_data);
			}
		}
#endif
		
	}

	int cmd = TrackPopupMenu(menu,TPM_RIGHTBUTTON|TPM_NONOTIFY|TPM_RETURNCMD,pt.x,pt.y,0,wnd,0);

	DestroyMenu(menu);

	if (cmd >= ID_OTHER && cmd < ID_EXT_MENU_BASE)
	{
		insert_extension(extensions[cmd-ID_OTHER].guid);
	}
	else if(cmd >= ID_EXT_MENU_BASE){
		if(ext.is_valid()){
			ext->get_menu_items(menu_hook);
		}
		menu_hook.execute_by_id(cmd);
	}
	else
	{
		switch(cmd)
		{
		case ID_CLOSE:
			if (target_tab != -1)
				delete_tab(target_tab);
			break;
		case ID_RENAME:
			if (target_tab != -1)
				do_tab_rename(target_tab);
			break;
		case ID_LEFT:
			if (target_tab != -1)
				swap_tabs(target_tab-1,target_tab);
			break;
		case ID_RIGHT:
			if (target_tab != -1)
				swap_tabs(target_tab,target_tab+1);
			break;
		case ID_TAB_TOP:
			set_tab_bar_orientation(TABS_ON_TOP);
			break;
		case ID_TAB_BOTTOM:
			set_tab_bar_orientation(TABS_ON_BOTTOM);
			break;
		case ID_TAB_LEFT:
			set_tab_bar_orientation(TABS_ON_LEFT);
			break;
#ifdef TAB_STYLE_CHANGE
		case ID_TAB_TAB:
			set_tab_style(TABS_TAB);
			break;
		case ID_TAB_BTN:
			set_tab_style(TABS_BUTTON);
			break;
#endif
        case ID_HIDETAB:
            {
                toggle_hide_tab();
            }
            break;
        case ID_LIMIT_MIN_HEIGHT:
            {
                change_limit_min_height();
            }
            break;
		case ID_SETTINGS:
			{
				static_api_ptr_t<ui_control> ui;
				ui->show_preferences(guid_config);
			}
			break;
/*		case ID_CLOSEPANEL:
			if (wnd && parent_ext && parent_ext->get_host().is_valid())
			{
				TRACK_CODE("ref add", parent_ext->service_add_ref());
				TRACK_CODE("hide", ShowWindow(wnd,SW_HIDE));
				SetParent(wnd,0);
				TRACK_CODE("relinquish_ownership", parent_ext->get_host()->relinquish_ownership(wnd));
				TRACK_CODE("destroy", destroy_window());
				TRACK_CODE("ref release", parent_ext->service_release());
			}
			break;
*/
		}
	}
}

HWND uieh_tabs::get_child_window(unsigned idx)
{
	ui_extension::window_ptr ext = get_child_extension(idx);
	if (ext.is_valid())
		return ext->get_wnd();
	return 0;
}

RECT uieh_tabs::get_display_area()
{
	RECT rc;
	GetClientRect(wnd,&rc);
	if (IsRectEmpty(&rc)) { SetRectEmpty(&rc); return rc; }

    if (!hide_tab) {
        TabCtrl_AdjustRect(tab_wnd, FALSE, &rc);
    	InflateRect(&rc,4*WINDOW_BORDER_SIZE_X,2*WINDOW_BORDER_SIZE_Y);
        if (get_tab_bar_orientation() == TABS_ON_TOP)
	        OffsetRect(&rc,0,2*WINDOW_BORDER_SIZE_Y);
        else if (get_tab_bar_orientation() == TABS_ON_BOTTOM)
	        OffsetRect(&rc,0,-2*WINDOW_BORDER_SIZE_Y);
    }
//	else
//		OffsetRect(&rc,0,2*WINDOW_BORDER_SIZE_Y);

	return rc;
}

void uieh_tabs::size_child_window(HWND child_wnd)
{
	if (!child_wnd || !IsChild(wnd,child_wnd)) return;
	
	RECT rc = get_display_area();
	SetWindowPos(child_wnd,0,
		rc.left, 
		rc.top, 
		rc.right - rc.left,
		rc.bottom - rc.top, 
		SWP_NOZORDER
		);
}

void uieh_tabs::on_tab_sel_change()
{
	selected_tab = TabCtrl_GetCurSel(tab_wnd);
	on_tab_sel_change(selected_tab);
}

void uieh_tabs::on_tab_sel_change(unsigned int p_index)
{
	selected_tab = p_index;
    unsigned n = exts.get_count();
	HDWP dwp = BeginDeferWindowPos(n);
    if (p_index != (unsigned)-1)
	{
		active_wnd = get_child_window(p_index);
		if (active_wnd)
		{
			RECT rc = get_display_area();
			dwp = DeferWindowPos(dwp,active_wnd,HWND_TOP,
				rc.left,
				rc.top,
				rc.right - rc.left,
				rc.bottom - rc.top,
				SWP_SHOWWINDOW
				);
		}
	}
	for (unsigned i = 0; i < n; i++)
		if (exts[i]->ptr->get_wnd() != active_wnd)
			dwp = DeferWindowPos(dwp,exts[i]->ptr->get_wnd(),0,0,0,0,0,
				SWP_HIDEWINDOW|SWP_NOMOVE|/*SWP_NOSIZE|*/SWP_NOZORDER|SWP_NOACTIVATE);
	                                      //for foo_browser
	EndDeferWindowPos(dwp);
}

bool uieh_tabs::is_visible(HWND wnd) const
{
	if (wnd == active_wnd && parent_ext->get_host()->is_visible(this->wnd))
		return true;
	return false;
}

bool uieh_tabs::is_visibility_modifiable(HWND child_wnd, bool desired_visibility) const
{
	if (!IsChild(wnd,child_wnd)) return false;

	if (desired_visibility)
	{
		ui_extension::window_host_ptr host = parent_ext->get_host();
		if (host.is_valid() && !host->is_visible(wnd) && !host->is_visibility_modifiable(wnd,true))
			return false;
		return true;
	}
	else
	{
		return parent_ext->get_host()->is_visibility_modifiable(wnd,false);
	}
}

bool uieh_tabs::set_window_visibility(HWND child_wnd, bool visibility)
{
	if (!IsChild(wnd,child_wnd)) return false;

	if (visibility)
	{
		if (child_wnd != active_wnd)
		{
			unsigned i, n = exts.get_count();
			for (i = 0; i < n; i++)
			{
				if (child_wnd == exts[i]->ptr->get_wnd())
				{
					TabCtrl_SetCurSel(tab_wnd,i);
					on_tab_sel_change();
					break;
				}
			}
		}
		if (!parent_ext->get_host()->is_visible(wnd))
			return parent_ext->get_host()->set_window_visibility(wnd,true);
		return true;
	}
	else
	{
		return parent_ext->get_host()->set_window_visibility(wnd,visibility);
	}
}

ui_extension::window_ptr& uieh_tabs::get_child_extension(unsigned idx) const
{
	if(!exts.get_count())
			return pre_exts[idx - exts.get_count()]->ptr;
	TCITEM item;
	memset(&item,0,sizeof(item));
	item.mask = TCIF_PARAM;
	TabCtrl_GetItem (tab_wnd,idx,&item);
	return reinterpret_cast<tabbed_ext*>(item.lParam)->ptr;
}

void uieh_tabs::swap_tabs(unsigned i1, unsigned i2)
{
	if(!exts.get_count()){
		pre_exts.swap_items(i1,i2);
		return;
	}

	unsigned max = TabCtrl_GetItemCount(tab_wnd);
	if (max == 0 || i1 == i2 || i1 >= max || i2 >= max){
		return;
	}

	TCITEM item1, item2;
	array_t<wchar_t> name1, name2;
	name1.set_size(30);
	name2.set_size(30);

	memset(&item1,0,sizeof(item1));
	item1.mask = TCIF_PARAM|TCIF_TEXT;
	item1.pszText = name1.get_ptr();
	item1.cchTextMax = 30;
	TabCtrl_GetItem(tab_wnd,i1,&item1);

	memset(&item2,0,sizeof(item2));
	item2.mask = TCIF_PARAM|TCIF_TEXT;
	item2.pszText = name2.get_ptr();
	item2.cchTextMax = 30;
	TabCtrl_GetItem(tab_wnd,i2,&item2);

	TabCtrl_SetItem(tab_wnd,i1,&item2);
	TabCtrl_SetItem(tab_wnd,i2,&item1);
	exts.swap_items(i1,i2);

	if (selected_tab == i1)
	{
		selected_tab = i2;
		TabCtrl_SetCurSel(tab_wnd,i2);
	}
	else if (selected_tab == i2)
	{
		selected_tab = i1;
		TabCtrl_SetCurSel(tab_wnd,i1);
	}
}

void uieh_tabs::delete_tab(unsigned idx, bool destroy_window)
{
	if(!exts.get_count()){
		pre_exts.delete_by_idx(idx);
		return ;
	}

	TCITEM item;
	memset(&item,0,sizeof(item));
	item.mask = TCIF_PARAM;
	TabCtrl_GetItem (tab_wnd,idx,&item);
	tabbed_ext * ext = reinterpret_cast<tabbed_ext*>(item.lParam);
	if (ext)
	{
		if(ext->ptr->get_wnd()){
			ShowWindow(ext->ptr->get_wnd(), SW_HIDE);
		}
		TabCtrl_DeleteItem(tab_wnd,idx);
		if (ext->ptr->get_wnd() == active_wnd)
		{
			active_wnd = 0;
			int new_sel = ((int)idx)==TabCtrl_GetItemCount(tab_wnd)? idx-1 : idx;
			TabCtrl_SetCurSel(tab_wnd,new_sel);
		}
		if (destroy_window)
			ext->ptr->destroy_window();
		exts.delete_item(ext);
		on_tab_sel_change();
	}
}

void uieh_tabs::rename_tab(unsigned idx, const char * new_name)
{

	if(!exts.get_count()){
		pre_exts[idx]->caption = new_name;
		return ;
	}

	uTCITEM item;
	memset(&item,0,sizeof(item));
	item.mask = TCIF_TEXT;
	item.pszText = const_cast<char *>(new_name);
	if (uTabCtrl_SetItem(tab_wnd,idx,&item))
	{
		assert(idx >= 0 && idx < exts.get_count());
		exts[idx]->caption = new_name;
		// in case any tabs have wrapped/unwrapped
		size_child_window(active_wnd);
	}
}

void uieh_tabs::do_tab_rename(unsigned idx)
{
	if (idx >= 0 && idx < exts.get_count())
	{
		string8 name = exts[idx]->caption;
		if (uDialogBox(IDD_RENAME,wnd,rename_proc,(long)&name) > 0)
			rename_tab(idx,name.get_ptr());
	}
}

BOOL CALLBACK uieh_tabs::rename_proc(HWND wnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		assert(lp != 0);
		uSetWindowLong(wnd,DWL_USER,lp);
		uSetDlgItemText(wnd,IDC_EDIT,reinterpret_cast<string_base*>(lp)->get_ptr());
		uSendDlgItemMessage(wnd,IDC_EDIT,EM_SETSEL,0,-1);
		SetFocus(GetDlgItem(wnd,IDC_EDIT));
		break;
	case WM_COMMAND:
		switch(wp)
		{
		case IDOK:
			uGetDlgItemText(wnd,IDC_EDIT,*reinterpret_cast<string_base*>(uGetWindowLong(wnd,DWL_USER)));
			EndDialog(wnd,1);
			break;
		case IDCANCEL:
			EndDialog(wnd,0);
			break;
		}
		break;
	}
	return 0;
}

BOOL CALLBACK uieh_tabs::set_limit_min_height_proc(HWND wnd, UINT msg, WPARAM wp, LPARAM lp)
{
    switch (msg) {
    case WM_INITDIALOG:
        uSetWindowText(wnd, "Changes minimum height");
        break;
    }
    return rename_proc(wnd, msg, wp, lp);
}

void uieh_tabs::set_tab_bar_orientation(tab_orientation o)
{
	if (get_tab_bar_orientation() != o)
	{
		long styles  = uGetWindowLong(tab_wnd,GWL_STYLE);
		switch(o)
		{
		case TABS_ON_TOP:
			styles &= ~(TCS_BOTTOM | TCS_VERTICAL);
			break;
		case TABS_ON_BOTTOM:
			styles |= TCS_BOTTOM;
			styles &= ~TCS_VERTICAL;
			break;
		case TABS_ON_LEFT:
			styles |= TCS_VERTICAL;
			styles &= ~TCS_BOTTOM;
			break;
		}
		uSetWindowLong(tab_wnd,GWL_STYLE,styles);
		SetWindowPos(tab_wnd,0,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOZORDER|SWP_FRAMECHANGED);
		uSendMessage(wnd,WM_MOVE,0,0);
		on_tab_sel_change();
	}
}

void uieh_tabs::set_tab_style(tab_style s)
{
	if (get_tab_style() != s)
	{
		long styles  = uGetWindowLong(tab_wnd,GWL_STYLE);
		switch(s)
		{
		case TABS_TAB:
			styles &= ~(TCS_BUTTONS | TCS_FLATBUTTONS);
//			styles |= TCS_MULTILINE;
			break;
		case TABS_BUTTON:
			styles |= TCS_BUTTONS | TCS_FLATBUTTONS;
//			styles &= ~(TCS_MULTILINE);
			break;
		}
		uSetWindowLong(tab_wnd,GWL_STYLE,styles);
//		uSendMessage(wnd,WM_MOVE,0,0);
	}
}

void uieh_tabs::toggle_hide_tab()
{
    hide_tab = !hide_tab;
    if (hide_tab) {
        ShowWindow(tab_wnd, SW_HIDE);
    } else {
        ShowWindow(tab_wnd, SW_SHOW);
    }
    get_display_area();
    if (exts.get_count()>0)
	rename_tab(0,exts[0]->caption);
}

void uieh_tabs::change_limit_min_height()
{
    string8 name = pfc::format_uint(limit_min_height);
    if (uDialogBox(IDD_RENAME,wnd,set_limit_min_height_proc,(long)&name) > 0) {
        if (pfc::string_is_numeric(name)) {
            limit_min_height = atoui_ex(name, 5);
        }
    }
    if (parent_ext && parent_ext->get_host().is_valid())
			parent_ext->get_host()->on_size_limit_change(wnd,ui_extension::size_limit_minimum_height);
}

void uieh_tabs::get_config(stream_writer * p_writer, abort_callback & p_abort)
{
	p_writer->write_string("tab_orientation", p_abort);
	if(tab_orient != (unsigned)-1)
		p_writer->write_lendian_t((DWORD)tab_orient, p_abort);
	else
		p_writer->write_lendian_t((DWORD)get_tab_bar_orientation(), p_abort);

	p_writer->write_string("selected_tab", p_abort);
	if(selected_tab_cfg != (unsigned)-1)
		p_writer->write_lendian_t(selected_tab_cfg, p_abort);
	else
		p_writer->write_lendian_t(selected_tab, p_abort);

    p_writer->write_string("hide_tab", p_abort);
	p_writer->write_lendian_t(static_cast<unsigned int>(hide_tab), p_abort);

    p_writer->write_string("limit_min_height", p_abort);
	p_writer->write_lendian_t(limit_min_height, p_abort);

	p_writer->write_string("ext_data", p_abort);
	DWORD n = exts.get_count();
	DWORD m = pre_exts.get_count();
	p_writer->write_lendian_t(n+m, p_abort);
	for (unsigned i = 0; i < n; i++)
	{
		p_writer->write_object_t(exts[i]->ptr->get_extension_guid(), p_abort);
		p_writer->write_string(exts[i]->caption, p_abort);

		stream_writer_memblock mem_writer;
		exts[i]->ptr->get_config(&mem_writer, p_abort); 

		DWORD mem_size = mem_writer.m_data.get_size();
		p_writer->write_lendian_t(mem_size, p_abort);
		if(mem_size > 0)
			p_writer->write(mem_writer.m_data.get_ptr(), mem_size, p_abort);
	}	
	for (unsigned i = 0; i < m; i++)
	{
		p_writer->write_object_t(pre_exts[i]->ptr->get_extension_guid(), p_abort);
		p_writer->write_string(pre_exts[i]->caption, p_abort);

		stream_writer_memblock mem_writer;
		pre_exts[i]->ptr->get_config(&mem_writer, p_abort); 

		DWORD mem_size = mem_writer.m_data.get_size();
		p_writer->write_lendian_t(mem_size, p_abort);
		if(mem_size > 0)
			p_writer->write(mem_writer.m_data.get_ptr(), mem_size, p_abort);
	}
}

void uieh_tabs::set_config(stream_reader * p_reader, t_size p_size, abort_callback & p_abort)
{
#if 1
	if (!p_reader || !p_size) return;
	DWORD sel_tab;
	string8_fastalloc tmp;
	stream_reader_limited_ref buf_reader(p_reader, p_size);
	for(;;)
	{
		buf_reader.read_string(tmp, p_abort);
		{
			if (!uStringCompare(tmp,"tab_orientation"))
			{
				DWORD orient;
				buf_reader.read_lendian_t(orient, p_abort);
				tab_orient = orient;
			}
			else if (!uStringCompare(tmp,"selected_tab")){
				buf_reader.read_lendian_t(sel_tab, p_abort);
				selected_tab_cfg = sel_tab;
			}
            else if (!uStringCompare(tmp,"hide_tab")) {
                unsigned int hide_tab_cfg;
                buf_reader.read_lendian_t(hide_tab_cfg, p_abort);
                hide_tab = static_cast<bool>(hide_tab_cfg);
            }
            else if (!uStringCompare(tmp,"limit_min_height")) {
                unsigned int limit_min_height_cfg;
                buf_reader.read_lendian_t(limit_min_height_cfg, p_abort);
                limit_min_height = limit_min_height_cfg;
            }
			else if (!uStringCompare(tmp,"ext_data")){
				break;
			}
			else{
				if(buf_reader.get_remaining())
					buf_reader.flush_remaining(p_abort);
				return;
			}
		}
	}
	DWORD num_exts = 0;
	buf_reader.read_lendian_t(num_exts, p_abort);

	for (unsigned i = 0; i < num_exts; i++)
	{
		GUID guid;
		buf_reader.read_object_t(guid, p_abort);

		string8 caption;
		buf_reader.read_string(caption, p_abort);

		DWORD config_size = 0;
		buf_reader.read_lendian_t(config_size, p_abort);

		stream_reader_limited_ref config_reader(&buf_reader, config_size);
		TRACK_CODE("insert ext", insert_extension(guid, &config_reader, config_size, p_abort, caption));
		if(config_reader.get_remaining())
			config_reader.flush_remaining(p_abort);

	}		
	if(buf_reader.get_remaining())		
		buf_reader.flush_remaining(p_abort);
#endif
}

bool uieh_tabs::override_status_text_create(service_ptr_t<ui_status_text_override> & p_out)
{
	if (parent_ext && parent_ext->get_host().is_valid())
		return parent_ext->get_host()->override_status_text_create(p_out);
	return false;
}

bool uieh_tabs::get_keyboard_shortcuts_enabled() const
{
	if (parent_ext && parent_ext->get_host().is_valid())
		return parent_ext->get_host()->get_keyboard_shortcuts_enabled();
	return false;
}


unsigned uieh_tabs::is_resize_supported(HWND child_wnd)
{
#if 1
	if (parent_ext && parent_ext->get_host().is_valid())
		return parent_ext->get_host()->is_resize_supported(wnd);
#endif
	return 0;
}

bool uieh_tabs::request_resize(HWND child_wnd, unsigned flags, unsigned width, unsigned height)
{
#if 1
	if (parent_ext && parent_ext->get_host().is_valid())
		return parent_ext->get_host()->request_resize(wnd,flags,width,height);
#endif
	return false;
}

uie_tabs_panel::uie_tabs_panel() : p_host(0)
{
	p_containing_host = new service_impl_t<uieh_tabs>(this);
}

uie_tabs_panel::~uie_tabs_panel()
{
	p_containing_host.release();
}

HWND uie_tabs_panel::create_or_transfer_window(HWND parent,const ui_extension::window_host_ptr &host,
							   const ui_helpers::window_position_t &t){
	HWND wnd = p_containing_host->get_wnd();
	if (wnd)
	{
		ShowWindow(wnd, SW_HIDE);
		SetParent(wnd, parent);
		if (p_host.is_valid())
		{
			p_host->relinquish_ownership(wnd);
			p_host=0;
		}
		p_host = host;
	}
	else
	{
		p_host = host;
		wnd = p_containing_host->create_window(parent);
	}
	return wnd;

}
void uie_tabs_panel::destroy_window() 
{
	if (p_containing_host.is_valid())
		TRACK_CODE("destroy panel", p_containing_host->destroy_window());
}

HWND uie_tabs_panel::get_wnd() const
{
	if (p_containing_host.is_valid())
		return p_containing_host->get_wnd();
	return 0;
}

void uie_tabs_panel::get_config(stream_writer * p_writer, abort_callback & p_abort) const
{
	if (p_containing_host.is_valid()){
		TRACK_CODE("write config", p_containing_host->get_config(p_writer, p_abort));
	}
}
void uie_tabs_panel::set_config(stream_reader * p_reader, t_size p_size, abort_callback & p_abort)
{
	if (p_containing_host.is_valid()){
		TRACK_CODE("read config", p_containing_host->set_config(p_reader, p_size, p_abort));
	}
}

void uie_tabs_panel::insert_panel(t_size index, const splitter_item_t * p_item){
	unsigned index2 = p_containing_host->get_child_count();
	bool custom_title = false;
	string8 tmp;
	const splitter_item_full_impl_t *p = 0;
	if(p_item->query(p) && p)
	{
		custom_title = p->m_custom_title;
		if(custom_title)
			p->get_title(tmp);
	} else {
		custom_title = false;
	}

	if(custom_title)
		p_containing_host->insert_extension(p_item->get_panel_guid(), tmp);
	else
		p_containing_host->insert_extension(p_item->get_panel_guid());

	if(index < index2)
		p_containing_host->swap_tabs(index, index2);

	stream_writer_memblock mem_writer;
	abort_callback_impl p_abort;
	p_item->get_panel_config(&mem_writer);
	if(mem_writer.m_data.get_size()){
		stream_reader_memblock_ref	
			mem_reader(mem_writer.m_data.get_ptr(), mem_writer.m_data.get_size());
		p_containing_host->get_child_extension(index)->set_config(&mem_reader, mem_writer.m_data.get_size(), p_abort);
	}

}

void uie_tabs_panel::remove_panel(t_size index){
	p_containing_host->delete_tab(index);
}
void uie_tabs_panel::replace_panel(t_size index, const splitter_item_t * p_item)
{
	unsigned index2 = p_containing_host->get_child_count();
	insert_panel(index2, p_item);
	if(index < index2){
		p_containing_host->swap_tabs(index, index2);
		p_containing_host->delete_tab(index2);
	}
}
t_size uie_tabs_panel::get_panel_count()const{
	return p_containing_host->get_child_count();
}


	template <class t_base>
	class splitter_item_simple_custom : public t_base
	{
	public:
		virtual const GUID & get_panel_guid()const {return m_guid;}
		virtual void get_panel_config (stream_writer * p_out)const 
		{
			abort_callback_impl p_abort;
			p_out->write(m_data.get_ptr(), m_data.get_size(), p_abort);
		}
		virtual void set_panel_guid(const GUID & p_guid) 
		{
			if (m_guid != p_guid)
			{
				m_guid = p_guid;
				m_data.set_size(0);
			}
		}
		virtual void set_panel_config (stream_reader * p_reader, t_size p_size) 
		{

			abort_callback_impl p_abort;
			m_data.set_size(p_size);
			p_reader->read(m_data.get_ptr(), m_data.get_size(), p_abort);
		}
		virtual const window_ptr & get_window_ptr()const{
			return m_ptr;
		}
		void set_window_ptr(const window_ptr & p_source){m_ptr = p_source;}

		inline splitter_item_simple_custom(const splitter_item_t & p_source) 
		{
			set(p_source);
		}
		inline splitter_item_simple_custom() : m_guid(pfc::guid_null)
		{
		}
		virtual bool query(const GUID & p_guid) const {
			return false;}
	protected:
		GUID m_guid;
		pfc::array_t<t_uint8> m_data;
		window_ptr m_ptr;
	};

	typedef splitter_item_simple_custom<splitter_item_t> splitter_item_simple_custom_t;

splitter_item_t * uie_tabs_panel::get_panel(t_size index)const
{
	splitter_item_full_impl_t *p = new splitter_item_full_impl_t();
//	splitter_item_simple_custom<splitter_item_t> *p = new splitter_item_simple_custom<splitter_item_t>;
	uie::window_ptr e = 
		p_containing_host->get_child_extension(index);	

	stream_writer_memblock mem_writer;
	abort_callback_impl abort;
	e->get_config(&mem_writer, abort);

	stream_reader_memblock_ref	
		mem_reader(mem_writer.m_data.get_ptr(), mem_writer.m_data.get_size());

	p->set_window_ptr(p_containing_host->get_child_extension(index));
	p->set_panel_guid(p_containing_host->get_child_extension(index)->get_extension_guid());
	p->set_panel_config(&mem_reader, mem_writer.m_data.get_size());
	p->m_custom_title = !p_containing_host->has_default_name(index);
	string8 name;
	p_containing_host->get_child_caption(index, name);
	p->set_title(name, name.length());
	e.release();
	return p;
}


/*
// {54F7E8E9-33EC-448c-ACCC-961C61F19025}
const GUID uie_tabs_panel::extension_guid = 
{ 0xa1ffc77e, 0x9b69, 0x44db, { 0x9a, 0xdc, 0x13, 0xea, 0xef, 0x94, 0x3a, 0x9 } };

// {59ADBF35-999B-4406-8173-3B57427435A7}
const GUID uieh_tabs::host_guid = 
{ 0xba1360f, 0xdb3b, 0x4de5, { 0x80, 0xf6, 0x2b, 0x14, 0x7c, 0x5b, 0x23, 0xb2 } };
// {A1FFC77E-9B69-44db-9ADC-13EAEF943A09}
// {0BA1360F-DB3B-4de5-80F6-2B147C5B23B2}
*/
// {E3956EA0-2BC8-4edf-914F-1209943BDD51}
const GUID uie_tabs_panel::extension_guid = 
{ 0xe3956ea0, 0x2bc8, 0x4edf, { 0x91, 0x4f, 0x12, 0x9, 0x94, 0x3b, 0xdd, 0x51 } };
// {F1375FBF-DF6B-4d78-A063-80DDF50A22A0}
const GUID uieh_tabs::host_guid = 
{ 0xf1375fbf, 0xdf6b, 0x4d78, { 0xa0, 0x63, 0x80, 0xdd, 0xf5, 0xa, 0x22, 0xa0 } };

void get_available_extension_infos(ui_extension::window_info_list_simple & out, DWORD type, ui_extension::window_host_ptr &host)
{
	string8_fastalloc tmp;
	service_enum_t<ui_extension::window> e;
	service_ptr_t<ui_extension::window> ptr;
	if(e.first(ptr)){
		do{
			if (ptr.is_valid() && (ptr->get_type() & type) && ptr->is_available(host))
			{
				ui_extension::window_info_simple info;
				info.guid = ptr->get_extension_guid();
				ptr->get_name(tmp);
				info.name = tmp;
				ptr->get_category(tmp);
				info.category = tmp;
				out.add_item(info);
			}
		}while(e.next(ptr));
	}
}
